/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 *
 * $FreeBSD: src/sys/kern/kern_jail.c,v 1.6.2.3 2001/08/17 01:00:26 rwatson Exp $
 * $DragonFly: src/sys/kern/kern_jail.c,v 1.5 2003/06/26 02:17:45 dillon Exp $
 *
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/sysproto.h>
#include <sys/kern_syscall.h>
#include <sys/malloc.h>
#include <sys/proc.h>
#include <sys/jail.h>
#include <sys/namei.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <sys/vnode.h>
#include <net/if.h>
#include <netinet/in.h>

MALLOC_DEFINE(M_PRISON, "prison", "Prison structures");

SYSCTL_NODE(, OID_AUTO, jail, CTLFLAG_RW, 0,
    "Jail rules");

int	jail_set_hostname_allowed = 1;
SYSCTL_INT(_jail, OID_AUTO, set_hostname_allowed, CTLFLAG_RW,
    &jail_set_hostname_allowed, 0,
    "Processes in jail can set their hostnames");

int	jail_socket_unixiproute_only = 1;
SYSCTL_INT(_jail, OID_AUTO, socket_unixiproute_only, CTLFLAG_RW,
    &jail_socket_unixiproute_only, 0,
    "Processes in jail are limited to creating UNIX/IPv4/route sockets only");

int	jail_sysvipc_allowed = 0;
SYSCTL_INT(_jail, OID_AUTO, sysvipc_allowed, CTLFLAG_RW,
    &jail_sysvipc_allowed, 0,
    "Processes in jail can use System V IPC primitives");

struct prisonlist prisons;
int jail_id_last = 0;
int prisoncount = 0;

static struct prison	*jail_find(int id);

static void
init_prison(void *data __unused)
{
	LIST_INIT(&prisons);
}

SYSINIT(prison, SI_SUB_INTRINSIC, SI_ORDER_ANY, init_prison, NULL);


/*
 * jail()
 *
 * jail_args(syscallarg(struct jail *) jail)
 */
int
jail(struct jail_args *uap) 
{
	int error;
	struct prison *pr;
	struct jail j;
	struct chroot_args ca;
	struct thread *td = curthread;
	struct proc *p = td->td_proc;
	struct nameidata nd;

	error = suser(td);
	if (prisoncount > PRISON_MAXID)
		error = EAGAIN;
	if (error)
		return (error);
	error = copyin(uap->jail, &j, sizeof j);
	if (error)
		return (error);
	if (j.version != 0)
		return (EINVAL);
	MALLOC(pr, struct prison *, sizeof *pr , M_PRISON, M_WAITOK);
	bzero((caddr_t)pr, sizeof *pr);
	error = copyinstr(j.hostname, &pr->pr_host, sizeof pr->pr_host, 0);
	if (error) 
		goto bail;
	pr->pr_ip = j.ip_number;
	error = copyinstr(j.path, &pr->pr_path, sizeof pr->pr_path, 0);
	if (error) 
		goto bail;
	pr->pr_version = PRISON_VERSION;
	pr->pr_linux = NULL;
	pr->pr_securelevel = securelevel;

		/*
		 * Find a free prison ID by trying one ID higher than the last
		 * ID we've used.  If that is already taken, try the next one
		 * higher until we've looped through all possible IDs.
		 */
	if (++jail_id_last > PRISON_MAXID)
		jail_id_last = 1;
	pr->pr_id = jail_id_last;

	while (jail_find(pr->pr_id) != NULL) {
		pr->pr_id++;
		if (pr->pr_id > PRISON_MAXID)
			pr->pr_id = 1;
		if (pr->pr_id == jail_id_last) {
			error = EAGAIN;
			goto bail;
		}
	}

		/*
		 * Store the jail's root vnode reference
		 */
	NDINIT(&nd, NAMEI_LOOKUP, CNP_FOLLOW | CNP_LOCKLEAF, UIO_SYSSPACE, pr->pr_path, td);
	error = namei(&nd);
	if (error)
		goto bail;
	pr->pr_root = nd.ni_vp;
	VOP_UNLOCK(nd.ni_vp, NULL, 0, td);
	NDFREE(&nd, NDF_ONLY_PNBUF);

	ca.path = j.path;
	error = chroot(&ca);
	if (error) {
		vrele(pr->pr_root);
		goto bail;
	}

	pr->pr_ref++;
	prisoncount++;
	LIST_INSERT_HEAD(&prisons, pr, pr_list);
	cratom(&p->p_ucred);
	p->p_ucred->cr_prison = pr;
	p->p_flag |= P_JAILED;
	return (0);

bail:
	FREE(pr, M_PRISON);
	return (error);
}

/*
 * jail_attach()
 *
 * jail_args(syscallarg(int) id)
 *
 * Attach the current process to the jail identified by id by chrooting
 * into the jail's root directory.
 */
int
jail_attach(struct jail_attach_args *uap) {
	int error, id = uap->id;
	struct thread *td = curthread;
	struct proc *p = td->td_proc;
	struct prison *pr;

	error = suser(td);
	if (error)
		return (error);
	pr = jail_find(id);
	if (pr == NULL)
		return (EINVAL);

	VOP_LOCK(pr->pr_root, NULL, LK_SHARED, td);
	error = kern_chroot(pr->pr_root);
	VOP_UNLOCK(pr->pr_root, NULL, 0, td);
	if (error)
		return (error);

	cratom(&p->p_ucred);
	setsugid();
	p->p_ucred->cr_prison = pr;
	p->p_flag |= P_JAILED;
	pr->pr_ref++;

	return (0);
}


/*
 * Returns a prison instance, or NULL on failure
 */
static struct prison *
jail_find(int id) {
	struct prison *pr;

	LIST_FOREACH(pr, &prisons, pr_list) {
		if (pr->pr_id == id)
			return (pr);
	}
	return (NULL);
}

int
prison_ip(struct thread *td, int flag, u_int32_t *ip)
{
	u_int32_t tmp;
	struct prison *pr;

	if (td->td_proc == NULL)
		return (0);
	if ((pr = td->td_proc->p_ucred->cr_prison) == NULL)
		return (0);
	if (flag) 
		tmp = *ip;
	else
		tmp = ntohl(*ip);
	if (tmp == INADDR_ANY) {
		if (flag) 
			*ip = pr->pr_ip;
		else
			*ip = htonl(pr->pr_ip);
		return (0);
	}
	if (tmp == INADDR_LOOPBACK) {
		if (flag)
			*ip = pr->pr_ip;
		else
			*ip = htonl(pr->pr_ip);
		return (0);
	}
	if (pr->pr_ip != tmp)
		return (1);
	return (0);
}

void
prison_remote_ip(struct thread *td, int flag, u_int32_t *ip)
{
	u_int32_t tmp;
	struct prison *pr;

	if (td == NULL || td->td_proc == NULL)
		return;
	if ((pr = td->td_proc->p_ucred->cr_prison) == NULL)
		return;
	if (flag)
		tmp = *ip;
	else
		tmp = ntohl(*ip);
	if (tmp == INADDR_LOOPBACK) {
		if (flag)
			*ip = pr->pr_ip;
		else
			*ip = htonl(pr->pr_ip);
	}
	return;
}

int
prison_if(struct thread *td, struct sockaddr *sa)
{
	struct prison *pr;
	struct sockaddr_in *sai = (struct sockaddr_in*) sa;
	int ok;

	if (td->td_proc == NULL)
		return(0);
	pr = td->td_proc->p_ucred->cr_prison;

	if ((sai->sin_family != AF_INET) && jail_socket_unixiproute_only)
		ok = 1;
	else if (sai->sin_family != AF_INET)
		ok = 0;
	else if (pr->pr_ip != ntohl(sai->sin_addr.s_addr))
		ok = 1;
	else
		ok = 0;
	return (ok);
}

static int
sysctl_jail_list(SYSCTL_HANDLER_ARGS)
{
	int count, error;
	struct prison *pr, *xp, *sxp;
	struct thread *td = req->td;
	struct proc *p = td->td_proc;

		/*
		 * Do not allow prisons to list other prisons
		 */
	if (p->p_ucred->cr_prison)
		return (0);
	count = prisoncount;
	if (count == 0)
		return (0);
	if (count < 0)
		panic("prisoncount %d < 0", count);

		/*
		 * Copy each prison structure and export them to userland
		 */
	sxp = xp = malloc(sizeof(*xp) * count, M_TEMP, M_WAITOK | M_ZERO);
	LIST_FOREACH(pr, &prisons, pr_list)
		bcopy(pr, xp++, sizeof(*xp));

	error = SYSCTL_OUT(req, sxp, sizeof(*sxp) * count);
	free(sxp, M_TEMP);
	if (error)
		return (error);
	return (0);
}

SYSCTL_OID(_jail, OID_AUTO, list, CTLTYPE_STRUCT | CTLFLAG_RD,
    NULL, 0, sysctl_jail_list, "S", "List of active jails");
